
import logging

logger = logging.getLogger('waspy')


class ErrorLoggingBase:
    def log_exception(self, request, exc_info, level='error'):
        try:
            level = getattr(logging, level.upper())
        except:
            level = logging.ERROR

        logger.log(level, 'An error occurred while handling request: {}'
                   .format(request),
                   exc_info=exc_info)

    def log_warning(self, request, message):
        logger.warning(message)


class SentryLogging(ErrorLoggingBase):
    def __init__(self, *, dsn, environment):
        try:
            from raven import Client
        except ImportError as e:
            raise ImportWarning(
                'You must install raven in order to use the sentry logger'
            ) from e

        self.raven = Client(dsn, environment=environment)
        self.raven.transaction.clear()

    def _get_sentry_details(self, request, exc_info):
        data = {
            'request': {
                'method': request.method.value,
                'data': request.body,
                'query_string': request.query_string,
                'url': '/' + request.path.replace('.', '/'),
                'content-type': request.content_type,
                'headers': request.headers
            },
            'user': {
            }
        }
        tags = {  # put things you want to filter by in sentry here
            'handler': request._handler.__qualname__,
        }
        extra_data = {  # put extra data here
            'correlation_id': request.correlation_id
        }

        self.add_context_data(request, data, tags, extra_data, exc_info)
        return data, tags, extra_data

    def log_warning(self, request, message=None):
        super().log_warning(request, message=message)
        data, tags, extra = self._get_sentry_details(request, None)
        self.raven.captureMessage(message, data=data, extra=extra, tags=tags)

    def log_exception(self, request, exc_info, level='error'):
        super().log_exception(request, exc_info, level=level)

        data, tags, extra = self._get_sentry_details(request, exc_info)
        self.raven.captureException(data=data, extra=extra, tags=tags,
                                    exc_info=exc_info, level=level)

    def add_context_data(self, request, data, tags, extra_data, exc_info):
        """
        Override this method to add your own sentry data.
        :param request: the request the error occured on
        :param data: This is the normal sentry data, prepopulated with the
        request data for you. Dictionary.
        :param tags: things you want to filter by in sentry. Dictionary
        :param extra_data: extra informational data. Dictionary
        :param exc_info: the exception info tuple provide from sys.exc_info()
        :return: Nothing. Just modify the passed in dictionaries
        """
        pass



import asyncio
import socket

import h11

from ..webtypes import Request, Response, ResponseError
from .transportabc import TransportABC, ClientTransportABC


class ClosedError(Exception):
    """ Error for closed connections """


class _HTTPClientConnection:
    slots = ('reader', 'writer', 'conn', 'port', 'service')

    def __init__(self, service, port):
        self.reader = None
        self.writer = None
        self.port = port
        self.service = service
        self.conn = h11.Connection(our_role=h11.CLIENT)

    async def connect(self):
        try:
            self.reader, self.writer = await \
                asyncio.open_connection(self.service, self.port)
        except socket.gaierror as e:
            raise ResponseError(e.strerror,
                                reason='Can not connect to service {service}.'
                                       .format(service=self.service),
                                status=503)

    def send(self, event):
        data = self.conn.send(event)
        if data is not None:
            self.writer.write(data)

    async def next_event(self):
        while True:
            event = self.conn.next_event()
            if event is h11.NEED_DATA:
                data = await self.reader.read(1064)
                self.conn.receive_data(data)

            else:
                return event

    def close(self):
        self.send(h11.ConnectionClosed())
        self.writer.close()


class HTTPClientTransport(ClientTransportABC):
    """Client implementation of the HTTP transport protocol"""
    def _get_connection_for_service(self, service):
        pass

    async def make_request(self, service, method, path, body=None, query=None,
                           headers=None, correlation_id=None,
                           content_type=None, port=80, **kwargs):
        # form request object
        path = path.replace('.', '/')
        if headers is None:
            headers = {}
        headers['Host'] = service
        headers['Connection'] = 'close'
        if correlation_id:
            headers['X-Correlation-Id'] = str(correlation_id)
        if query:
            path += '?' + query
        if content_type:
            headers['Content-Type'] = content_type
        if body:
            headers['Content-Length'] = str(len(body))
        else:
            headers['Content-Length'] = '0'

        # now make a connection and send it
        connection = _HTTPClientConnection(service, port)
        await connection.connect()
        connection.send(h11.Request(method=method, target=path,
                                    headers=headers.items()))
        if body:
            connection.send(h11.Data(data=body))

        connection.send(h11.EndOfMessage())

        response = await connection.next_event()
        assert type(response) is h11.Response

        # form response object
        status_code = response.status_code
        headers = {name.decode('ascii'): value.decode('ascii')
                   for name, value in response.headers}

        result = Response(headers=headers, correlation_id=correlation_id,
                          status=status_code)

        body = b''
        event = None
        while type(event) is not h11.EndOfMessage:
            event = await connection.next_event()
            if type(event) is h11.Data:
                body += event.data

        result.body = body
        connection.close()
        return result


class HTTPTransport(TransportABC):
    """ Server implementation of the HTTP transport protocol"""

    def get_client(self):
        return HTTPClientTransport()

    def __init__(self, port=8080, prefix=None, shutdown_grace_period=15):
        """
        HTTP Transport for listening on http
        :param port: The port to lisen on (0.0.0.0 will always be used)
        :param prefix: the path prefix to remove from all url's
        :param shutdown_grace_period: Time to wait for server to shutdown
        before connections get forceably closed. The only way for connections
        to not be forcibly closed is to have some connection draining in front
        of the service for deploys. Most docker schedulers will do this for you.
        """
        self.port = port
        if prefix is None:
            prefix = ''
        self.prefix = prefix
        self.shutdown_grace_period = shutdown_grace_period
        self._handler = None
        self._server = None
        self._done_future = asyncio.Future()
        self._count = 0
        self._shutting_down = False
        self._sleeping_connections = set()

    def listen(self, *, loop: asyncio.AbstractEventLoop):
        coro = asyncio.start_server(
            self.handle_incoming_request, '0.0.0.0', self.port, loop=loop)
        self._server = loop.run_until_complete(coro)
        print('-- Listening for HTTP on port {} --'.format(self.port))

    async def start(self, request_handler):
        self._handler = request_handler
        try:
            await self._done_future
        except asyncio.CancelledError:
            pass

        # Enter shutdown step

        print('shutting down http')
        self._shutting_down = True
        for coro in self._sleeping_connections:
            coro.cancel()

        # we should wait some time for connections to stop
        previous_count = self._count
        for i in range(self.shutdown_grace_period):
            await asyncio.sleep(1)
            if self._count == previous_count:
                # no more connections
                break
            else:
                previous_count = self._count

        self._server.close()
        await self._server.wait_closed()

    async def handle_incoming_request(self, reader, writer):
        if self._shutting_down:
            self._count += 1
        self._set_tcp_nodelay(writer)
        protocol = _HTTPServerProtocol(reader=reader, writer=writer,
                                       prefix=self.prefix)
        try:
            while True:
                inner = asyncio.Task(protocol.get_request())
                coro = asyncio.shield(inner)
                self._sleeping_connections.add(coro)
                try:
                    request = await coro
                except asyncio.CancelledError:
                    # Give the request a chance to send a request if it has one
                    if inner.done():
                        request = inner.result()
                    else:
                        request = await asyncio.wait_for(inner, 1)
                self._sleeping_connections.remove(coro)
                response = await self._handler(request)
                await protocol.send_response(response)
                if protocol.conn.our_state is h11.MUST_CLOSE:
                    break
                if self._shutting_down:
                    break
                protocol.conn.start_next_cycle()
        except (ClosedError, ConnectionResetError,
                asyncio.CancelledError, asyncio.TimeoutError):
            pass
        finally:
            writer.close()

    def shutdown(self):
        self._done_future.cancel()

    def _set_tcp_nodelay(self, writer):
        socket_ = writer._transport._sock
        if socket_ is None:
            return
        if socket_.family not in (socket.AF_INET, socket.AF_INET6):
            return
        socket_.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, True)


class _HTTPServerProtocol(asyncio.Protocol):
    """ HTTP Protocol handler.
        Should only be used by HTTPServerTransport
    """
    __slots__ = ('conn', 'reader', 'writer')

    def __init__(self, reader=None, writer=None, prefix=None):
        self.conn = h11.Connection(h11.SERVER)
        self.reader = reader
        self.writer = writer
        self.prefix = prefix

    async def _read(self):
        if self.conn.they_are_waiting_for_100_continue:
            go_ahead = h11.InformationalResponse(status_code=100)
            await self.writer.write(go_ahead)
        try:
            data = await self.reader.read(1000000)
        except ConnectionError:
            data = b''
        self.conn.receive_data(data)

    async def send_response(self, response: Response):
        headers = response.headers
        if response.content_type:
            headers['content-type'] = response.content_type
        if response.correlation_id:
            headers['X-Correlation-Id'] = response.correlation_id
        if response.data:
            headers['content-length'] = str(len(response.data))
        else:
            headers['content-length'] = '0'
        headers = [(name, value) for name, value in headers.items()]

        r = h11.Response(status_code=response.status.value, headers=headers,
                         reason=response.status.phrase)
        await self._send(r)
        data = response.data
        if data is not None:
            await self._send(h11.Data(data=data))
        await self._send(h11.EndOfMessage())

    async def _send(self, r):
        data = self.conn.send(r)
        self.writer.write(data)
        await self.writer.drain()

    async def get_request(self) -> Request:

        event = await self.next_event()
        if isinstance(event, h11.ConnectionClosed):
            raise ClosedError
        headers = {name.decode('ascii'): value.decode('ascii')
                   for name, value in event.headers}

        # split path and query string
        target = event.target.decode('ascii').split('?', maxsplit=1)
        path = target[0].lstrip(self.prefix)
        try:
            query = target[1]
        except IndexError:
            query = None
        method = event.method.decode('ascii')

        body = b''

        while True:
            event = await self.next_event()
            if type(event) is h11.EndOfMessage:
                break
            body += event.data
        correlation_id = headers.get('x-correlation-id')
        return Request(
            headers=headers,
            method=method,
            query_string=query,
            path=path,
            body=body,
            correlation_id=correlation_id
        )

    async def next_event(self):
        while True:
            event = self.conn.next_event()
            if event is h11.NEED_DATA:
                await self._read()
            else:
                return event



from setuptools import setup

try:
    import pypandoc
    long_description = pypandoc.convert('readme.md', 'rst')
except(IOError, ImportError):
    long_description = open('readme.md').read()

setup(
    name='waspy',
    version='0.13.14',
    install_requires=[
        'h11==0.7.0',
        'aioamqp==0.8.2',
        'pyyaml==3.12',
        'aenum==1.4.5'
    ],
    packages=['waspy', 'waspy.transports'],
    long_description=long_description,
    url='https://github.com/wasp/waspy',
    license='Apache 2.0',
    author='nhumrich',
    author_email='nick at humrich dot us',
    description='Async Microservices Framework',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='wasp async asyncio curio rest framework rabbitmq microservices'
)





import sys
import asyncio
import signal
from typing import List, Union, Iterable
from http import HTTPStatus
from concurrent.futures import CancelledError

from .client import Client
from .webtypes import Request, Response, ResponseError
from .router import Router
from .transports.transportabc import TransportABC
from .configuration import Config, ConfigError
from . import errorlogging


async def response_wrapper_factory(app, handler):
    async def wrap_response_middleware(request):
        response = await handler(request)
        if not isinstance(response, Response):
            if isinstance(response, tuple):
                body = response[0]
                status = response[1]
                response = Response(status=status, body=body)
            elif isinstance(response, dict) or isinstance(response, str):
                response = Response(body=response)
            elif response is None:
                response = Response(status=HTTPStatus.NO_CONTENT)
            else:
                raise ValueError('Request handler returned an invalid type.'
                                 ' Return types should be one of '
                                 '[Response, dict, str, None, (dict, int)]')
        return response
    return wrap_response_middleware


class Application:
    def __init__(self,
                 transport: Union[TransportABC,
                                  Iterable[TransportABC]]=None,
                 *,
                 middlewares: List[callable]=None,
                 default_headers: dict=None,
                 debug: bool=False,
                 router: Router=None,
                 config: Config=None):
        if transport is None:
            from waspy.transports.httptransport import HTTPTransport
            transport = HTTPTransport()
        if not isinstance(transport, tuple):
            transport = (transport,)
        if middlewares is None:
            middlewares = ()
        middlewares = tuple([m for m in middlewares])
        middlewares += (response_wrapper_factory,)
        if router is None:
            router = Router()
        if default_headers is None:
            default_headers = {'Server': 'waspy'}
        if not config:
            config = Config()
        self.transport = transport
        self.middlewares = middlewares
        self.default_headers = default_headers
        self.debug = debug
        self.router = router
        self.on_start = []
        self._client = None
        self.config = config
        self.raven = None
        self.logger = None

    @property
    def client(self) -> Client:
        if not self._client:
            self._client = Client(transport=self.transport[0].get_client())
        return self._client

    def start_shutdown(self, signum=None, frame=None):
        print('Starting Shutdown')
        # loop = asyncio.get_event_loop()
        for t in self.transport:
            t.shutdown()

    def run(self):
        # first, lets make the signals kill us

        loop = asyncio.get_event_loop()
        # init logger
        self._create_logger()

        # wrap handlers in middleware
        loop.run_until_complete(self._wrap_handlers())
        for t in self.transport:
            t.listen(loop=loop)

        # Call on-startup hooks
        loop.run_until_complete(asyncio.gather(*[
            coro(self) for coro in self.on_start
        ]))

        # todo: fork/add processes?
        tasks = []
        for t in self.transport:
            tasks.append(t.start(self.handle_request))

        # register signals, so that stopping the service works correctly
        loop.add_signal_handler(signal.SIGTERM, self.start_shutdown)
        loop.add_signal_handler(signal.SIGINT, self.start_shutdown)

        # Run all transports - they shouldn't return until shutdown
        loop.run_until_complete(asyncio.gather(*tasks))

        # todo: Call on-shutdown hooks

        # exit

    async def handle_request(self, request: Request) -> Response:
        """
        coroutine: This method is called by Transport
        implementation to handle the actual request.
        It returns a webtype.Response object.
        """
        # Get handler
        try:
            try:
                handler = self.router.get_handler_for_request(request)
                request.app = self
                response = await handler(request)

            except ResponseError as r:
                response = r.response
                if r.log:
                    exc_info = sys.exc_info()
                    self.logger.log_exception(request, exc_info, level='warning')
            # invoke serialization (json) to make sure it works
            _ = response.data

        except CancelledError:
            # This error can happen if a client closes the connection
            # The response shouldnt really ever be used
            return None
        except Exception:
            exc_info = sys.exc_info()
            self.logger.log_exception(request, exc_info)
            response = Response(status=500,
                                body={'message': 'Server Error'})
        if not response.correlation_id:
            response.correlation_id = request.correlation_id

        # add default headers
        response.headers = {**self.default_headers, **response.headers}

        return response

    async def _wrap_handlers(self):
        handler_gen = self.router._get_and_wrap_routes()

        try:
            handler = next(handler_gen)
            while True:
                wrapped = handler
                for middleware in self.middlewares[::-1]:
                    wrapped = await middleware(self, wrapped)
                handler = handler_gen.send(wrapped)
        except StopIteration:
            pass

    def _create_logger(self):
        try:
            dsn = self.config['sentry']['dsn']
        except (ConfigError, ValueError):
            self.logger = errorlogging.ErrorLoggingBase()
        else:
            try:
                env = self.config['app_env']
            except (ConfigError):
                env = 'waspy'
            self.logger = errorlogging.SentryLogging(
                dsn=dsn,
                environment=env
            )

